#Imports
import sys
import os
import signal
import time
import argparse
import json
from queue import Queue, Empty
from gpiozero.pins.pigpio import PiGPIOFactory
from gpiozero import Device, LED
import pigpio
import logging
import logging.handlers
import BME280
from Hologram.HologramCloud import HologramCloud

#Globals
logger = None

#Main Application Class
class RemotePreHeat:
    stop = False

    #updates the environmental data
    def _updateEnvData(self):
        result = self._bme280.read_data()
        self._envdata['temperature'], self._envdata['pressure'], self._envdata['humidity'] = result
        logger.debug("Environemnt data read: Temperature: {}°C, Pressure: {}Pa, Humidity: {}%".format(*result))
        return result

    def __init__(self, altitude=0):
        #Instance Variable declaration
        self._hologram = None
        self._bme280 = None
        self._envdata = dict(temperature=0, humidity=0, pressure=0)
        self._PWR = [None] * 2
        self._PWRLED = [None] * 2
        self._msgQueue = Queue()
        self.altitude = altitude

        logger.info("Initializing system...")

        #First, let's initialize the GPIOs
        logger.debug("Initializing GPIOs")
        pio = pigpio.pi()
        Device.pin_factory = PiGPIOFactory()
        logger.debug("Pins used: PWR0: 4, PWR1: 17, LED0: 18, LED 1: 6")
        self._PWR[0] = LED(4, active_high=False)
        self._PWR[1] = LED(17, active_high=False)
        self._PWRLED[0] = LED(18)
        self._PWRLED[1] = LED(6)

        #Second, let's setup the BME280 sensor
        logger.debug("Initialize BME280 Temperature/Humidity/Pressure sensor")
        self._bme280 = BME280.sensor(pio, interface=BME280.SPI)
        self._updateEnvData()

        #Now, let's initialize the modem.
        #Deactivate modem power saving features
        logger.debug("Initializing HologramCloud")
        self._hologram = HologramCloud(dict(), network='cellular')
        if(self._hologram is None):
            extxt = "Could not create modem interface"
            logger.critical(extxt, exc_info=True)
            raise Exception(extxt)
        
        try:
            self._hologram.network.modem.command('+CFUN', '0')
            #self._hologram.network.modem.command('+CPSMS', '0')
            self._hologram.network.modem.command('+CEDRXS', '0,2')
            self._hologram.network.modem.command('+CEDRXS', '0,4')
            self._hologram.network.modem.command('+CEDRXS', '0,5')
            self._hologram.network.modem.command('+CFUN', '15')
        except Exception as e:
            logger.critical("Deactivating modem power saving features failed.", exc_info=True)
            raise e

        #Wait for the modem to come back
        del self._hologram
        self._hologram = None
        modemQueryTimeout = 10
        while( (self._hologram is None) and (modemQueryTimeout > 0) ):
            try:
                logger.debug(f"Reinitializing HologramCloud after modem restart, timeout {modemQueryTimeout}")
                self._hologram = HologramCloud(dict(), network='cellular')
            except:
                self._hologram = None
                modemQueryTimeout -= 1
                time.sleep(1)
        
        if(modemQueryTimeout == 0):
            extxt = "Could not reconnect to modem after disabling power saving features"
            logger.critical(extxt, exc_info=True)
            raise Exception(extxt)

        #wait for Network
        modemQueryTimeout = 30
        while( (self._hologram.network.operator == None) and (modemQueryTimeout > 0) ):
            logger.debug(f"Waiting for MNO connection, timeout {modemQueryTimeout}")
            time.sleep(1)
            modemQueryTimeout -= 1

        if(modemQueryTimeout == 0):
            extxt = "Could not connect to MNO"
            logger.critical(extxt, exc_info=True)
            raise Exception(extxt)
        else:
            logger.debug(f"Connected to {self._hologram.network.operator}")
        
        self._hologram.enableSMS()



#Main loop
    def Run(self):
        ltime = time.time()

        #Open the receive socket to the hologram cloud
        logger.info("Starting listener...")
        try:
            logger.debug("Connecting to HologramCloud...\nopening socket connection and subscribing to the message factory.")
            self._hologram.openReceiveSocket()
        except Exception as e:
            logger.critical(e, exc_info=True)
            raise e

        #Loop di Loop
        logger.debug("Starting main loop...")
        self._hologram.network.modem.command('+CEDRXS?')
        while(not RemotePreHeat.stop):
            msg = None

            try:
                msg = self._hologram.popReceivedMessage()
                if(msg is None):
                    time.sleep(1)
                    if(time.time() - ltime > 30):
                        self._hologram.network.modem.command('+USOCTL', '0,10')
                        self._hologram.network.modem.command('+CREG?')
                        #self._hologram.sendMessage("", topics=["HB"])
                        ltime = time.time()
                    continue
                else:
                    logger.debug("Received message: " + msg)
            except Exception as e:
                logger.exception("Something weird failed in the message queue, trying to continue...\n{}".format(e), exc_info=True)
                continue

            msgData = json.loads(msg)

            if(msgData['type'] == 'status'):
                logger.debug("Cloud requested a status update...")
                self._updateEnvData()
                mbPressure = float(self._envdata["pressure"]) / 100
                altimeterSetting = (mbPressure - 0.3) * pow(1 + ((pow(1013.25, 0.190284) * 0.0065 / 288.0) * (self.altitude / pow(mbPressure - 0.3, 0.190284))), 1.0 / 0.190284)
                returnMsg = json.dumps(
                    {
                        "Temperature": self._envdata["temperature"],
                        "Humidity": self._envdata["humidity"],
                        "Pressure": altimeterSetting,
                        "Outlet1": self._PWR[0].value,
                        "Outlet2": self._PWR[1].value
                    }
                )
                logger.debug("Sending return message:\n{}".format(returnMsg))
                try:
                    self._hologram.sendMessage(returnMsg, topics=["sendMessage"])
                except Exception as e:
                    logger.exception("Something weird failed sending a cloud message, trying to continue...\n{}".format(e), exc_info=True)

            elif(str(msgData['type']).startswith("Outlet")):
                outletNum = int(str(msgData['type'])[-1]) - 1
                onoff = bool(msgData['active'])
                logger.debug("Cloud requested to turn Outlet {} {}".format(outletNum, "on" if onoff else "off"))
                if(onoff):
                    self._PWR[outletNum].on()
                    self._PWRLED[outletNum].on()
                else:
                    self._PWR[outletNum].off()
                    self._PWRLED[outletNum].off()


        #Clean up...
        logger.debug("Cleaning up...\nunsubscribing from message factory and closing socket connection to the HologramCloud,.")
        self._hologram.closeReceiveSocket()




#SIGINT handler to exit the program gracefully
def sigint_handler(sig, frame):
    logger.info("Program will be exited due to CTRL+C SIGINT")
    RemotePreHeat.stop = True

if __name__ == "__main__":
    os.system('clear')

    #Firs, let's parse the command line
    parser = argparse.ArgumentParser(description="Remote PreHeat\nCellular Power Outlet Controller")
    parser.add_argument("-altitude", "-a", action='store', default=0, metavar="ALT [m]", help="Sets the altitude the device is operated at for pressure calculations")
    parser.add_argument("-debug", "-d", action='store_true', help="Enable debug output")
    parser.add_argument("-verbose", "-v", action='store_true', help="Enable verbose log output")
    args = parser.parse_args()

    #Second, let's initialize the logging facility
    loggerSDK = logging.getLogger('')
    if(args.debug):
        loggerSDK.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")
    loggerSDK.addHandler(handler)

    logger = logging.getLogger('Remote PreHeat')
    if(args.verbose):
        logger.setLevel(logging.INFO)
    elif(args.debug):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)
    handler = logging.handlers.SysLogHandler(address = '/dev/log', facility=logging.handlers.SysLogHandler.LOG_LOCAL1)
    formatter = logging.Formatter('Python: { "loggerName":"%(name)s", "timestamp":"%(asctime)s", "pathName":"%(pathname)s", "logRecordCreationTime":"%(created)f", "functionName":"%(funcName)s", "levelNo":"%(levelno)s", "lineNo":"%(lineno)d", "time":"%(msecs)d", "levelName":"%(levelname)s", "message":"%(message)s"}')
    handler.formatter = formatter
    logger.addHandler(handler)

    logger.debug("Starting Remote PreHeat...\nCMD Line arguments: {}".format(args))

    #Now, Hook up the SIGINT handler in case we want to exit
    signal.signal(signal.SIGINT, sigint_handler)
    
    #Finally, let's create the main class and run it.
    rph = None
    
    try:
        rph = RemotePreHeat(altitude=float(args.altitude))
        rph.Run()
    except Exception as e:
        print(f"An oopsie happened: {e}")
    finally:
        if(rph is not None):
            del rph
            rph = None

    exit(0)
